from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple, List, Sequence, Union

from pybi.core.components.reactiveComponent.echarts import EChartUpdateInfo
from pybi.core.dataSource import DataSourceTable

if TYPE_CHECKING:
    from pybi.core.components.reactiveComponent import EChartDatasetInfo

m_merge_keys = set(["legend", "title", "grid", "series"])


class BaseChart:
    def __init__(self) -> None:
        self.__base_opt = {
            "tooltip": {},
            "legend": {},
            "series": [],
            "title": {},
            "grid": {"containLabel": True},
        }
        self.__merge_opt = {}
        self._updateInfos: List[EChartUpdateInfo] = []

    def __add__(self, other: BaseChart):
        return ChartCollector().append(self).append(other)

    def merge(self, options: Dict):
        self.__merge_opt = options
        return self

    def get_options(self):
        return {
            **self.__base_opt,
            **{k: v for k, v in self.__merge_opt.items() if k in m_merge_keys},
        }

    def set_title(self, text: str):
        opt_title: Dict = self.__base_opt["title"]
        opt_title["text"] = text

        return self

    def hover_filter(
        self, value_type: str, table: Union[str, DataSourceTable], field: str
    ):
        """
        value_type: Literal["x", "y", "value","color","name"]
        """
        if isinstance(table, DataSourceTable):
            table = table.source_name

        self._updateInfos.append(EChartUpdateInfo("hover", value_type, table, field))

        return self

    def click_filter(
        self, value_type: str, table: Union[str, DataSourceTable], field: str
    ):
        """
        value_type: Literal["x", "y", "value","color","name"]
        """
        if isinstance(table, DataSourceTable):
            table = table.source_name

        self._updateInfos.append(EChartUpdateInfo("click", value_type, table, field))

        return self

    def _remove_filters(self, actionType: str):
        """
        actionType : 'click' | 'hover'
        """
        self._updateInfos = [
            info for info in self._updateInfos if info.actionType != actionType
        ]
        return self

    def remove_all_click_filter(self):
        return self._remove_filters("click")

    def _create_default_click_filter(self):
        pass

    def get_options_infos(
        self,
    ) -> Tuple[Dict, List[EChartUpdateInfo]]:
        raise NotImplementedError


class ChartCollector:
    def __init__(self) -> None:
        self._collector: List[BaseChart] = []

    def append(self, other: BaseChart):
        self._collector.append(other)
        return self

    def __add__(self, other: Union[ChartCollector, BaseChart]):
        if isinstance(other, BaseChart):
            other = ChartCollector().append(other)

        self._collector.extend(other._collector)
        return self



from __future__ import annotations
from .line import LineChart
from .bar import BarChart
from .pie import PieChart
from .scatter import ScatterChart
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


__all__ = ["easy_echarts"]


class EasyEChartsSettings:
    def __init__(self) -> None:
        self.drill_down_default_set_click_filter = True


class EasyEChartsMeta:
    def __init__(self) -> None:
        self._settings = EasyEChartsSettings()

    def off_drill_down_default_set_click_filter(self):
        """
        默认情况下,多个组合图表配置会自动为其每个图表添加 `click_filter` 联动。
        此函数调用后,会关闭其功能
        """
        self._settings.drill_down_default_set_click_filter = False

    def make_line(
        self,
        data: DataSourceTable,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
    ):
        return LineChart(data, x, y, color, agg)

    def make_bar(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
    ):
        return BarChart(data, x, y, color, agg)

    def make_pie(
        self,
        data: DataSourceTable,
        *,
        name: str,
        value: str,
        agg="round(avg(${}),2)",
    ):
        return PieChart(data, name, value, agg)

    def make_scatter(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="round(avg(${}),2)",
    ):
        return ScatterChart(data, x, y, color, agg)


easy_echarts = EasyEChartsMeta()



from __future__ import annotations
from typing import Dict, List, Union


def copt2opt_obj(opts: Dict, name: str):
    if name in opts:
        target = opts[name]

        if isinstance(target, list) and len(target) == 0:
            return

        target = target[0] if isinstance(target, list) else target
        target = {**target}
        return target

    return None


def try_del_prop(obj: Dict, name: str):
    if name in obj:
        del obj[name]


def merge_user_options(
    base_opts: Dict,
):
    def update_config(series_config: Dict):
        series = copt2opt_obj(base_opts, "series")
        if series:

            try_del_prop(series, "data")
            try_del_prop(series, "type")

            series_config.update(series)

    def update_axis(options: Dict):
        xAxis = copt2opt_obj(base_opts, "xAxis")
        if xAxis:
            try_del_prop(xAxis, "data")
        yAxis = copt2opt_obj(base_opts, "yAxis")
        if yAxis:
            try_del_prop(yAxis, "data")

        options["xAxis"][0].update(xAxis)
        options["yAxis"][0].update(yAxis)

    def remove_series():
        if "series" in base_opts:
            base_opts["series"] = []

        if "legend" in base_opts:
            if len(base_opts["legend"]) > 0:
                target = base_opts["legend"][0]
                if "data" in target:
                    del target["data"]

    return update_config, update_axis, remove_series



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Dict, Set, Any, Union

from pybi.utils.data_gen import Jsonable
from pybi.core.components import ComponentTag
from .base import ReactiveComponent
from pybi.core.dataSource import DataSourceTable


if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable
    from pybi.core.sql import SqlInfo


class EChartSqlInfo(Jsonable):
    def __init__(
        self, seriesConfig: Dict, path: str, sql: SqlInfo, chartType: str
    ) -> None:
        self.path = path
        self._sql = sql
        self.chartType = chartType
        self.seriesConfig = seriesConfig

    def _to_json_dict(self):
        data = super()._to_json_dict()
        sql_data = self._sql._to_json_dict()

        data = {**data, **sql_data}
        return data


class EChartDatasetInfo(Jsonable):
    def __init__(self, seriesConfig: Dict, path: str, sql_info: SqlInfo) -> None:
        self.path = path
        self.sqlInfo = sql_info
        self.seriesConfig = seriesConfig


class EChartUpdateInfo(Jsonable):
    def __init__(
        self,
        action_type: str,
        value_type: str,
        table: str,
        field: str,
    ) -> None:
        """
        action_type : Literal["hover", "click"]
        value_type: Literal["x", "y", "value","color","name"]
        """
        super().__init__()
        self.actionType = action_type
        self.valueType = value_type
        self.table = table
        self.field = field


class EChartInfo(Jsonable):
    def __init__(
        self,
        options: Dict,
        datasetInfos: List[EChartDatasetInfo],
        updateInfos: List[EChartUpdateInfo],
    ):
        self.options = options
        self.datasetInfos = datasetInfos
        self.updateInfos = updateInfos


class EChart(ReactiveComponent):
    def __init__(self, *, option_type="dict") -> None:

        super().__init__(ComponentTag.EChart)

        self._chart_mappings = {}
        # self._options = options

        self._chartInfos: List[EChartInfo] = []
        self.optionType = option_type

        # self.options = options
        # self._sqlInfos: List[EChartDatasetInfo] = []
        # self._chartUpdateInfos: List[EChartUpdateInfo] = []
        self.height = "100%"

    def set_height(self, value: str):
        """
        15em:15字体大小
        300px:300像素
        """
        self.height = value
        return self

    def _add_chart_info(self, info: EChartInfo):
        self._chartInfos.append(info)
        return self

    def hover_filter(
        self, value_type: str, table: Union[str, DataSourceTable], field: str
    ):
        """
        value_type: , Literal["x", "y", "value","color","name"]
        """
        if isinstance(table, DataSourceTable):
            table = table.source_name

        self._chartInfos[0].updateInfos.append(
            EChartUpdateInfo("hover", value_type, table, field)
        )

        return self

    def _to_json_dict(self):
        # self.options["series"] = []
        # EChart.remove_sql_from_option(self.options, self._sqlInfos)

        if len(self._chartInfos) == 1:
            options = self._chartInfos[0].options
            # EChart.remove_legend_empty_data(options)
            # EChart.remove_axis_empty_data(options)

        data = super()._to_json_dict()
        data["mapping"] = self._chart_mappings

        data["chartInfos"] = self._chartInfos

        return data

    # @staticmethod
    # def remove_legend_empty_data(options: Dict):
    #     target = options
    #     if "legend" in target:
    #         target = options["legend"]
    #         if isinstance(target, list) and len(target) > 0:
    #             target = target[0]
    #             if (
    #                 "data" in target
    #                 and isinstance(target["data"], list)
    #                 and len(target["data"]) == 0
    #                 or (not target["data"][0])
    #             ):
    #                 target["data"] = None

    # @staticmethod
    # def remove_axis_empty_data(options: Dict):
    #     def remove_fn(name: str):
    #         target = options
    #         if name not in target or target[name] is None:
    #             return

    #         if "data" in target[name][0]:
    #             axis = target[name][0]
    #             target = axis["data"]
    #             if (
    #                 isinstance(target, list)
    #                 and len(target) == 0
    #                 or (not target[0] == "")
    #             ):
    #                 axis["data"] = None
    #                 axis["type"] = "category"

    #     remove_fn("xAxis")
    #     remove_fn("yAxis")

    # @staticmethod
    # def extract_infos_from_option(opt: Dict):

    #     yield from EChart._extract_series_infos_from_option_(opt)
    #     return
    #     stack = [("", opt)]

    #     while len(stack) > 0:

    #         path, target = stack.pop()

    #         for key, value in target.items():
    #             if key != "series":
    #                 cur_path = f"{path}.{key}"

    #                 if isinstance(value, dict):
    #                     stack.append((cur_path, value))
    #                 elif isinstance(value, list):
    #                     for idx, v in enumerate(value):
    #                         if isinstance(v, dict):
    #                             stack.append((f"{cur_path}[{idx}]", v))
    #                 elif isinstance(value, (Sql, DataSourceView)):

    #                     if isinstance(value, DataSourceView):
    #                         value = Sql(value._to_sql())

    #                     yield EChartSqlInfo(cur_path[1:], value, "other")

    # @staticmethod
    # def _extract_series_infos_from_option_(opt: Dict):
    #     if "series" not in opt:
    #         return

    #     def extract_series(series: Dict, idx=-1):
    #         series_data = series["data"]
    #         if series["type"] in ["line", "pie", "bar"]:
    #             if isinstance(series_data, (Sql, DataSourceView)):
    #                 if isinstance(series_data, DataSourceView):
    #                     series_data = Sql(series_data._to_sql())

    #                 path = "series"
    #                 if idx > -1:
    #                     path = f"{path}[{idx}].data"

    #                 series["pybiFlag"] = None

    #                 yield EChartSqlInfo(series, path, series_data, series["type"])

    #     series = opt["series"]
    #     if isinstance(series, list):
    #         for idx, s in enumerate(series):
    #             yield from extract_series(s, idx)
    #     else:
    #         yield from extract_series(series)

    # @staticmethod
    # def _extract_dataset_infos_from_option_(opt: Dict):
    #     if "dataset" not in opt:
    #         return

    #     series = []
    #     if "series" in opt:
    #         series = opt["series"]

    #     def extract_ds(ds: Dict, series: List[Dict], idx=-1):
    #         if "source" not in ds:
    #             return

    #         source = ds["source"]

    #         if isinstance(source, (DataSourceTable)):

    #             path = "dataset"
    #             if idx > -1:
    #                 path = f"{path}[{idx}]"

    #             # series["pybiFlag"] = None

    #             if "data" in series[0]:
    #                 series[0]["data"] = None

    #             yield EChartDatasetInfo(series[0], path, source.source_name)

    #     dataset = opt["dataset"]
    #     assert isinstance(dataset, list)

    #     for idx, ds in enumerate(dataset):
    #         yield from extract_ds(ds, series, idx)

    # @staticmethod
    # def remove_sql_from_option(opt: Dict, infos: List[EChartDatasetInfo]):

    #     filterSeries = [series for series in opt["series"] if not "pybiFlag" in series]

    #     opt["series"] = filterSeries

    #     # for info in infos:
    #     #     paths = info.path.split(".")
    #     #     dictUtils.set_by_paths(paths, opt, None)



from __future__ import annotations
import pybi.utils.sql as sqlUtils
from pybi.utils.data_gen import Jsonable
from typing import Dict, Generator, Optional

import re

m_sql_from_text_pat = re.compile(
    r"(?:info:\[\|sql:)(.+?)\|type:(.+?)\|js_map:(.*?)(?:\|])", re.I | re.DOTALL
)


class SqlWrapper(Jsonable):
    def __init__(self, sql: str) -> None:
        super().__init__()
        self._sql_info = SqlInfo(sql)

    def split_group(self, by: str, code: str):
        code = f"""
function temp(key, rows, index){{
    {code}
}}
const gp = utils.groupBy(rows,'{by}')
const res =  Object.entries(gp).map(([key, rows], index) => temp(key,rows,index))
return res
        """
        return self.js_map(code)

    def js_map(self, code: str):
        """
        code: js code
            js vars:
                `rows` is object : query result rows,each row is object
                `fields` is string array : query result field names
                `utils` utils functions : `{uniq,groupBy}`
        ---
        pbi.sql(f"select name,age,value from {data}")
            `rows` e.g `[{name:'x1',age:20,value:100},{name:'x2',age:30,value:200},...]`
            `fields` e.g `['name','age','value']`
        """
        self._sql_info.jsMap = code
        self._sql_info.set_udf_type()
        return self

    def toflatlist(self):
        self.js_map(
            """
if (fields.length>1){
    return rows.map(data=> fields.map(f=> data[f]))
}else {
     return rows.map(data=>data[fields[0]])
}"""
        )
        return self

    def __str__(self) -> str:
        return str(self._sql_info)

    def _to_json_dict(self):
        return self._sql_info


class SqlInfo(Jsonable):
    def __init__(
        self, sql: str, type: str = "infer", js_map: Optional[str] = None
    ) -> None:
        """
        type:
            'udf' : user defined function
            'infer' : automatically extrapolate from the results
        """
        super().__init__()
        self.sql = sql
        self.type = type
        self.jsMap = js_map

    def set_udf_type(self):
        self.type = "udf"
        return self

    def __str__(self) -> str:
        return f"info:[|sql:{self.sql}|type:{self.type}|js_map:{self.jsMap or ''}|]"

    @staticmethod
    def extract_sql_from_text(text: str):
        """
        >>> input = '总销售额:sql:[_ select sum(销售额) from data _]'
        >>> extract_sql_from_text(input)
        >>> ['总销售额:',Sql('select sum(销售额) from data')]
        """
        start_idx = 0

        for match in re.finditer(m_sql_from_text_pat, text):

            span = match.span()

            if span[0] > start_idx:
                # 前面有普通文本
                yield text[start_idx : span[0]]

            js_map = match.group(3)
            yield SqlInfo(
                match.group(1), match.group(2), None if len(js_map) == 0 else js_map
            )
            start_idx = span[1]

        end_idx = len(text) - 1

        if start_idx < end_idx:
            yield text[start_idx : len(text)]


class SqlExtractor:
    @staticmethod
    def extract_from_dict(data: Dict) -> Generator[tuple[str, SqlWrapper], None, None]:
        stack = [(data, "")]

        while len(stack) > 0:
            target, path = stack.pop()

            if isinstance(target, dict):
                inputs = ((value, f"{path}.{key}") for key, value in target.items())
                stack.extend(inputs)
                continue

            if isinstance(target, list):
                inputs = ((value, f"{path}[{idx}]") for idx, value in enumerate(target))
                stack.extend(inputs)
                continue

            if isinstance(target, SqlWrapper):
                yield path[1:], target
